"""Definition of Kirin's Intermediate Representation (IR) nodes."""

from kirin.ir.nodes.base import IRNode as IRNode
from kirin.ir.nodes.stmt import Statement as Statement
from kirin.ir.nodes.block import Block as Block
from kirin.ir.nodes.region import Region as Region
